#include "Utils.h"
#include "Win32Wrapper.h"

CWin32Wrapper::CWin32Wrapper(NUtils::FileType fileType)
: m_Type(fileType), m_lpMusicInfo(NULL)
{
	m_vf_callbacks.read_func  = s_vorbisfile_read_func;
	m_vf_callbacks.seek_func  = s_vorbisfile_seek_func;
	m_vf_callbacks.tell_func  = s_vorbisfile_tell_func;
	m_vf_callbacks.close_func = s_vorbisfile_close_func;
}

CWin32Wrapper::~CWin32Wrapper()
{
}

BOOL CWin32Wrapper::Process(LPDVPFILEINFOMUSIC lpMusicInfo)
{
	switch(m_Type)
	{
	case NUtils::TYPE_OGGVORBIS:
		return processOggVorbis(lpMusicInfo);
	case NUtils::TYPE_FLAC:
	case NUtils::TYPE_OGGFLAC:
		return processFlac(lpMusicInfo);
	default:
	case NUtils::TYPE_UNKNOWN:
		return FALSE;
	}
}

BOOL CWin32Wrapper::processOggVorbis(LPDVPFILEINFOMUSIC lpMusicInfo)
{
	BOOL bResult = FALSE;

	OggVorbis_File vf;

	if (NUtils::TYPE_OGGVORBIS == m_Type
	&&	0 <= ov_open_callbacks(this, &vf, NULL, 0, m_vf_callbacks))
	{
		// Static stuff.

		NUtils::StringCopy(lpMusicInfo->lpszFormat, TEXT("Ogg Vorbis"), lpMusicInfo->cchFormatMax);

		// Vorbis Info stuff.

		vorbis_info *vi=ov_info(&vf,-1);

		if (vi)
		{
			lpMusicInfo->dwSampleRate = vi->rate;
			lpMusicInfo->iNumChannels = vi->channels;

			if (vi->bitrate_lower != vi->bitrate_nominal
			||	vi->bitrate_upper != vi->bitrate_nominal
			||	0 == vi->bitrate_nominal)
			{
				lpMusicInfo->dwMusicFlags |= DVPMusicFlag_VBRAccurate;
			}

			ogg_int64_t oi64Length = ov_pcm_total(&vf,-1);
			long lAverageBitrate = ov_bitrate(&vf,-1);

			// Length won't be available for non-seekable bitstreams.
			if (OV_EINVAL != oi64Length)
			{
				ogg_int64_t oi64Duration = oi64Length;
				oi64Duration /= vi->rate;
				lpMusicInfo->dwDuration = (DWORD)oi64Duration;
			}

			if (OV_EINVAL != lAverageBitrate && OV_FALSE != lAverageBitrate)
			{
				// Actual average bitrate of the file or nominal bitrate if the input cannot seek.
				lpMusicInfo->dwBitRate = lAverageBitrate;
			}
		}

		// Vorbis Comments stuff.

		vorbis_comment *pComment = ov_comment(&vf,-1);

		if (pComment)
		{
			NUtils::StringCopyConvert(lpMusicInfo->lpszEncoder, pComment->vendor, lpMusicInfo->cchEncoderMax);

			// Work out max buffer size for a comment and allocate it.
			int nMaxLength = 0;
			int *pLength = pComment->comment_lengths;
			for(int i = 0; i < pComment->comments; i++)
			{
				if (nMaxLength < pLength[i])
				{
					nMaxLength = pLength[i];
				}
			}

			char **ptr = pComment->user_comments;

			if (ptr && 0 < nMaxLength)
			{
				char *szBufferKey  = new char[nMaxLength];
				char *szBufferData = new char[nMaxLength];

				if (NULL != szBufferKey && NULL != szBufferData)
				{
					// This is built by the decoder library so we should be able to trust it to be properly terminated.
					while(*ptr)
					{
						parseComment(lpMusicInfo, *ptr++, szBufferKey, szBufferData, nMaxLength);
					}
				}

				delete [] szBufferKey;
				delete [] szBufferData;
			}
		}

		bResult = TRUE;

		ov_clear(&vf);
	}

	return bResult;
}

BOOL CWin32Wrapper::processFlac(LPDVPFILEINFOMUSIC lpMusicInfo)
{
	BOOL bResult = TRUE;

	if (NUtils::TYPE_FLAC == m_Type)
	{
		NUtils::StringCopy(lpMusicInfo->lpszFormat, TEXT("FLAC"), lpMusicInfo->cchFormatMax);
	}
	else if (NUtils::TYPE_OGGFLAC == m_Type)
	{
		NUtils::StringCopy(lpMusicInfo->lpszFormat, TEXT("Ogg FLAC"), lpMusicInfo->cchFormatMax);
	}
	else
	{
		bResult = FALSE;
	}

	if (bResult)
	{
		m_lpMusicInfo = lpMusicInfo;

		EasyFLAC__StreamDecoder *pFlacDecoder = EasyFLAC__stream_decoder_new( NUtils::TYPE_OGGFLAC == m_Type );

		EasyFLAC__set_client_data(pFlacDecoder, this);

		EasyFLAC__set_read_callback(pFlacDecoder, s_easyflac_read_func);
		EasyFLAC__set_write_callback(pFlacDecoder, s_easyflac_write_func);
		EasyFLAC__set_error_callback(pFlacDecoder, s_easyflac_error_func);

		EasyFLAC__set_metadata_callback(pFlacDecoder, s_easyflac_metadata_func);
		EasyFLAC__set_metadata_respond(pFlacDecoder, FLAC__METADATA_TYPE_STREAMINFO);
		EasyFLAC__set_metadata_respond(pFlacDecoder, FLAC__METADATA_TYPE_VORBIS_COMMENT);

		EasyFLAC__init(pFlacDecoder);

		// Metadata callback will process metadata.
		EasyFLAC__process_until_end_of_metadata(pFlacDecoder);

		EasyFLAC__finish(pFlacDecoder);
		EasyFLAC__stream_decoder_delete(pFlacDecoder);

		m_lpMusicInfo = NULL;
	}

	return bResult;
}

void CWin32Wrapper::m_easyflac_metadata_func(const EasyFLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata)
{
	switch (metadata->type)
	{
	case FLAC__METADATA_TYPE_STREAMINFO:
		if (NULL != m_lpMusicInfo)
		{
			m_lpMusicInfo->dwSampleRate = metadata->data.stream_info.sample_rate;

			ULARGE_INTEGER uliDuration;
			uliDuration.QuadPart = metadata->data.stream_info.total_samples;
			uliDuration.QuadPart /= metadata->data.stream_info.sample_rate;

			m_lpMusicInfo->dwDuration = uliDuration.LowPart;

			m_lpMusicInfo->iNumChannels = metadata->data.stream_info.channels;
		}
		break;

	case FLAC__METADATA_TYPE_VORBIS_COMMENT:
		// Unlike the nicer VorbisFile API, these strings are not terminated for us.

		if (NULL != m_lpMusicInfo)
		{
			NUtils::StringExtractConvert(m_lpMusicInfo->lpszEncoder,   metadata->data.vorbis_comment.vendor_string.entry,
											m_lpMusicInfo->cchEncoderMax, metadata->data.vorbis_comment.vendor_string.length);

			unsigned int numComments = metadata->data.vorbis_comment.num_comments;

			size_t nMaxLength = 0;

			for (unsigned int i = 0; i < numComments; i++)
			{
				if (nMaxLength < metadata->data.vorbis_comment.comments[i].length)
				{
					nMaxLength = metadata->data.vorbis_comment.comments[i].length;
				}
			}

			char *szBufferKey  = new char[nMaxLength];
			char *szBufferData = new char[nMaxLength];

			for (unsigned int i = 0; i < numComments; i++)
			{
				char *szExtracted = NUtils::StringExtract(metadata->data.vorbis_comment.comments[i].entry,
															metadata->data.vorbis_comment.comments[i].length);

				if (NULL != szExtracted)
				{
					parseComment(m_lpMusicInfo, szExtracted, szBufferKey, szBufferData, nMaxLength);

					delete [] szExtracted;
				}
			}

			delete [] szBufferKey;
			delete [] szBufferData;
		}
		break;

	default:
		break;
	}
}

void CWin32Wrapper::parseComment(LPDVPFILEINFOMUSIC lpMusicInfo, const char *szComment, char *szBufferKey, char *szBufferData, size_t nBufferSize)
{
	size_t nKeyLength = strcspn(szComment, "=");

	if (0 < nKeyLength && nKeyLength < nBufferSize)
	{
		strncpy(szBufferKey, szComment, nKeyLength);
		szBufferKey[nKeyLength] = '\0';

		size_t nDataLength = strlen(szComment + nKeyLength + 1);

		if (0 < nDataLength && nDataLength < nBufferSize)
		{
			strcpy(szBufferData, (szComment + nKeyLength + 1));

			if (0 == stricmp(szBufferKey, "TITLE"))
			{
				NUtils::StringCopyConvert(lpMusicInfo->lpszTitle, szBufferData, lpMusicInfo->cchTitleMax);
			}
			else if (0 == stricmp(szBufferKey, "ALBUM"))
			{
				NUtils::StringCopyConvert(lpMusicInfo->lpszAlbum, szBufferData, lpMusicInfo->cchAlbumMax);
			}
			else if (0 == stricmp(szBufferKey, "ARTIST"))
			{
				NUtils::StringCopyConvert(lpMusicInfo->lpszArtist, szBufferData, lpMusicInfo->cchArtistMax);
			}
			else if (0 == stricmp(szBufferKey, "GENRE"))
			{
				NUtils::StringCopyConvert(lpMusicInfo->lpszGenre, szBufferData, lpMusicInfo->cchGenreMax);
			}
			else if (0 == stricmp(szBufferKey, "COMMENT"))
			{
				NUtils::StringCopyConvert(lpMusicInfo->lpszComment, szBufferData, lpMusicInfo->cchCommentMax);
			}
			else if (0 == stricmp(szBufferKey, "TRACKNUMBER"))
			{
				lpMusicInfo->iTrackNum = atoi(szBufferData);
			}
			else if (0 == stricmp(szBufferKey, "DATE"))
			{
				lpMusicInfo->iYear = atoi(szBufferData); // This could be more intelligent.
			}
		}
	}
}